<?php

add_filter(
    'ethereum_wallet_settings_tabs',
    'ETHEREUM_WALLET_blockchain_settings_tabs_hook',
    30,
    1
);
function ETHEREUM_WALLET_blockchain_settings_tabs_hook( $possible_screens )
{
    $possible_screens['blockchain'] = esc_html( __( 'Blockchain', 'ethereum-wallet' ) );
    return $possible_screens;
}

add_filter(
    'ethereum_wallet_get_save_options',
    'ETHEREUM_WALLET_blockchain_get_save_options_hook',
    27,
    3
);
function ETHEREUM_WALLET_blockchain_get_save_options_hook( $new_options, $options, $current_screen )
{
    if ( 'blockchain' !== $current_screen ) {
        return $new_options;
    }
    // Blockchain options screen
    $new_options['blockchain_api_node'] = 'infuraio';
    // if (ethereum_wallet_freemius_init()->can_use_premium_code__premium_only()) {
    $new_options['etherscanApiKey'] = ( !empty($_POST['ETHEREUM_WALLET_etherscanApiKey']) ? sanitize_text_field( $_POST['ETHEREUM_WALLET_etherscanApiKey'] ) : (( !empty($new_options['etherscanApiKey']) ? $new_options['etherscanApiKey'] : '' )) );
    $new_options['gas_limit'] = ( !empty($_POST['ETHEREUM_WALLET_gas_limit']) && is_numeric( $_POST['ETHEREUM_WALLET_gas_limit'] ) ? intval( sanitize_text_field( $_POST['ETHEREUM_WALLET_gas_limit'] ) ) : 200000 );
    $new_options['gas_price'] = ( !empty($_POST['ETHEREUM_WALLET_gas_price']) && is_numeric( $_POST['ETHEREUM_WALLET_gas_price'] ) ? floatval( sanitize_text_field( $_POST['ETHEREUM_WALLET_gas_price'] ) ) : 200 );
    
    if ( 'infuraio' === $new_options['blockchain_api_node'] ) {
        $new_options['blockchain_network'] = ( !empty($_POST['ETHEREUM_WALLET_blockchain_network']) ? sanitize_text_field( $_POST['ETHEREUM_WALLET_blockchain_network'] ) : 'mainnet' );
        $new_options['infuraApiKey'] = ( !empty($_POST['ETHEREUM_WALLET_infuraApiKey']) ? sanitize_text_field( $_POST['ETHEREUM_WALLET_infuraApiKey'] ) : '' );
    }
    
    return $new_options;
}

add_filter(
    'ethereum_wallet_print_options',
    'ETHEREUM_WALLET_blockchain_print_options_hook',
    27,
    2
);
function ETHEREUM_WALLET_blockchain_print_options_hook( $options, $current_screen )
{
    if ( 'blockchain' !== $current_screen ) {
        return;
    }
    $blockchain_api_nodes = [
        'infuraio' => __( 'Infura.io', 'ethereum-wallet' ),
        'custom'   => __( 'Custom', 'ethereum-wallet' ),
    ];
    $blockchain_api_node_current = ( isset( $options['blockchain_api_node'] ) && !empty($options['blockchain_api_node']) ? $options['blockchain_api_node'] : (( isset( $options['web3Endpoint'] ) && !empty($options['web3Endpoint']) ? 'custom' : 'infuraio' )) );
    ?>
    <script type='text/javascript'>
        if ('undefined' === typeof window['ethereumWallet']) {
            window.ethereumWallet = {};
        }
        jQuery(document).ready(function() {
            jQuery('#ETHEREUM_WALLET_blockchain_api_node').on('change', function(e) {
                const blockchain_api_node = jQuery(this).find(":selected").val();
                const blockchain_api_node_class = '.ethereum-wallet-blockchain-api-node-' + blockchain_api_node;
                jQuery('.ethereum-wallet-blockchain-api-node').hide();
                jQuery(blockchain_api_node_class).show();
                switch (blockchain_api_node) {
                    case 'infuraio':
                        jQuery('#ETHEREUM_WALLET_blockchain_network').change();
                        window.ethereumWallet._url_to_network_id_call_success = true;
                        break;
                    case 'custom':
                        jQuery('#ETHEREUM_WALLET_web3Endpoint').change();
                        break;
                }
            }).change();
        });
    </script>

    <tr valign="top">
        <th scope="row"><?php 
    _e( "Blockchain API Node", 'ethereum-wallet' );
    ?></th>
        <td>
            <fieldset>
                <label>
                    <select name="ETHEREUM_WALLET_blockchain_api_node" id="ETHEREUM_WALLET_blockchain_api_node" class="select short">
                        <?php 
    if ( is_array( $blockchain_api_nodes ) ) {
        foreach ( $blockchain_api_nodes as $blockchain_api_node => $blockchain_api_node_title ) {
            $selected = '';
            if ( $blockchain_api_node === $blockchain_api_node_current ) {
                $selected = 'selected="selected"';
            }
            ?>
                            <option value="<?php 
            echo  esc_attr( $blockchain_api_node ) ;
            ?>" <?php 
            echo  $selected ;
            ?>><?php 
            echo  esc_html( $blockchain_api_node_title ) ;
            ?></option>
                        <?php 
        }
    }
    ?>
                    </select>
                    <p><?php 
    _e( "The blockchain API node used: infura.io or custom", 'ethereum-wallet' );
    ?></p>
                </label>
            </fieldset>
        </td>
    </tr>
<?php 
}

add_filter(
    'ethereum_wallet_print_options',
    'ETHEREUM_WALLET_blockchain_print_options_hook2',
    28,
    2
);
function ETHEREUM_WALLET_blockchain_print_options_hook2( $options, $current_screen )
{
    if ( 'blockchain' !== $current_screen ) {
        return;
    }
    // infura.io supported blockchains
    $blockchains = apply_filters( 'ethereumico.io/supported-blockchains', [] );
    // ETHEREUM_WALLET_log('ETHEREUM_WALLET_blockchain_print_options_hook2: blockchains = ' . print_r($blockchains, true));
    ?>
    <tr valign="top" class="ethereum-wallet-blockchain-api-node ethereum-wallet-blockchain-api-node-infuraio">
        <th scope="row"><?php 
    _e( "Blockchain", 'ethereum-wallet' );
    ?></th>
        <td>
            <fieldset>
                <label>
                    <select name="ETHEREUM_WALLET_blockchain_network" id="ETHEREUM_WALLET_blockchain_network" class="select short">
                        <?php 
    if ( is_array( $blockchains ) ) {
        foreach ( $blockchains as $blockchain => $blockchain_title ) {
            $selected = '';
            if ( $blockchain === $options['blockchain_network'] ) {
                $selected = 'selected="selected"';
            }
            ?>
                            <option value="<?php 
            echo  esc_attr( $blockchain ) ;
            ?>" <?php 
            echo  $selected ;
            ?>><?php 
            echo  esc_html( $blockchain_title ) ;
            ?></option>
                        <?php 
        }
    }
    ?>
                    </select>
                    <p><?php 
    _e( "The blockchain used: mainnet, goerli or sepolia. Use mainnet in production, and goerli or sepolia in test mode. See plugin documentation for the testing guide.", 'ethereum-wallet' );
    ?></p>
                    <p><strong><?php 
    _e( "NOTE: The Polygon, Optimism, Arbitrum and Avalanche C-Chain chains require special registration on the infura side.", 'ethereum-wallet' );
    ?></strong></p>
                </label>
            </fieldset>
        </td>
    </tr>

    <tr valign="top" class="ethereum-wallet-blockchain-api-node ethereum-wallet-blockchain-api-node-infuraio">
        <th scope="row"><?php 
    _e( "Infura.io API Key", 'ethereum-wallet' );
    ?><sup>*</sup></th>
        <td>
            <fieldset>
                <label>
                    <input class="text" name="ETHEREUM_WALLET_infuraApiKey" id="ETHEREUM_WALLET_infuraApiKey" type="text" maxlength="70" placeholder="<?php 
    _e( "Put your Infura.io API Key here", 'ethereum-wallet' );
    ?>" value="<?php 
    echo  ( !empty($options['infuraApiKey']) ? esc_attr( $options['infuraApiKey'] ) : '' ) ;
    ?>">
                    <p><?php 
    echo  sprintf(
        __( 'The API key for the %1$s. You need to register on this site to obtain it. Follow the %2$sGet infura API Key%3$s guide please.', 'ethereum-wallet' ),
        '<a target="_blank" href="https://infura.io/register">https://infura.io/</a>',
        '<a target="_blank" href="https://ethereumico.io/knowledge-base/infura-api-key-guide/">',
        '</a>'
    ) ;
    ?></p>
                    <p><strong>
                            <?php 
    echo  sprintf( __( 'Note that this setting is ignored if the "%1$s" setting is set', 'ethereum-wallet' ), __( "Ethereum Node JSON-RPC Endpoint", 'ethereum-wallet' ) ) ;
    ?></strong></p>
                </label>
            </fieldset>
        </td>
    </tr>
<?php 
}

add_filter(
    'ethereum_wallet_print_options',
    'ETHEREUM_WALLET_blockchain_print_options_hook3',
    31,
    2
);
function ETHEREUM_WALLET_blockchain_print_options_hook3( $options, $current_screen )
{
    if ( 'blockchain' !== $current_screen ) {
        return;
    }
    $blockchainInfos = apply_filters( 'ethereumico.io/supported-blockchains-info', [] );
    $chainId = ETHEREUM_WALLET_getChainId();
    $blockchainInfo = apply_filters( 'ethereumico.io/blockchain-info', null, $chainId );
    ETHEREUM_WALLET_log( 'ETHEREUM_WALLET_blockchain_print_options_hook3: chainId = ' . $chainId . '; ' );
    $template = __( 'The API key for the %1$s. You need to %2$sregister%3$s on this site to obtain it.', 'ethereum-wallet' );
    $explorer_api_url = ( $blockchainInfo ? $blockchainInfo['explorer_api_url'] : null );
    $explorer_register_url = ( $blockchainInfo ? $blockchainInfo['explorer_register_url'] : null );
    $base_explorer_display_url = ( $blockchainInfo ? $blockchainInfo['base_explorer_display_url'] : null );
    $blockchain_api_node_current = ( isset( $options['blockchain_api_node'] ) && !empty($options['blockchain_api_node']) ? $options['blockchain_api_node'] : (( isset( $options['web3Endpoint'] ) && !empty($options['web3Endpoint']) ? 'custom' : 'infuraio' )) );
    $failedNetworkIdRequestMessage = sprintf( __( "Failed to request blockchain network ID. Check your \"%s\" setting value.", 'ethereum-wallet' ), __( "Ethereum Node JSON-RPC Endpoint", 'ethereum-wallet' ) );
    ?>
    <script type='text/javascript'>
        if ('undefined' === typeof window['ethereumWallet']) {
            window.ethereumWallet = {};
        }
        if ('undefined' === typeof window.ethereumWallet['_url_to_network_id_cache']) {
            window.ethereumWallet._url_to_network_id_cache = {};
        }
        if ('undefined' === typeof window.ethereumWallet['_url_to_network_id_call_started']) {
            window.ethereumWallet._url_to_network_id_call_started = {};
        }

        window.ethereumWallet.ethereum_wallet_get_blockchain_network_id = function(web3Endpoint, blockchainInfos, callback) {
            let alerts = {};

            function process_alerts() {
                Object.keys(alerts).forEach(function(message) {
                    alert(message);
                });
            }

            if ('' == web3Endpoint) {
                callback.call(null, null, null, alerts);
                process_alerts();
                return;
            }
            if ('undefined' !== typeof window.ethereumWallet._url_to_network_id_cache[web3Endpoint]) {
                window.ethereumWallet._url_to_network_id_call_success = true;
                callback.call(null, null, window.ethereumWallet._url_to_network_id_cache[web3Endpoint], alerts);
                process_alerts();
                return;
            }
            if ('undefined' === typeof window.ethereumWallet._url_to_network_id_call_started[web3Endpoint]) {
                window.ethereumWallet._url_to_network_id_call_started[web3Endpoint] = [];
            }
            window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].push(callback);
            if (window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].length > 1) {
                return;
            }

            window.ethereumWallet._url_to_network_id_call_success = '<?php 
    echo  $failedNetworkIdRequestMessage ;
    ?>';

            if ('undefined' !== typeof window.ethereumWallet['_url_to_network_id_timeout_id'] &&
                null !== window.ethereumWallet._url_to_network_id_timeout_id
            ) {
                clearTimeout(window.ethereumWallet._url_to_network_id_timeout_id);
                window.ethereumWallet._url_to_network_id_timeout_id = null;
            }
            window.ethereumWallet._url_to_network_id_timeout_id = setTimeout(function() {
                if (null !== window.ethereumWallet._url_to_network_id_timeout_id) {
                    clearTimeout(window.ethereumWallet._url_to_network_id_timeout_id);
                    window.ethereumWallet._url_to_network_id_timeout_id = null;
                }
                const dataRequest = '{"jsonrpc":"2.0","method":"net_version","params":[],"id":1}';
                jQuery.ajax({
                    url: web3Endpoint,
                    type: "POST",
                    data: dataRequest,
                    contentType: "application/json; charset=utf-8",
                    dataType: "json",
                    success: function(data, textStatus, jqXHR) {
                        // data: {"jsonrpc":"2.0","id":1,"result":"56"}
                        if (!data) {
                            console.log('Empty data returned for network', web3Endpoint);
                            window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].forEach(function(callback) {
                                callback.call(null, 'Empty data returned for network ' + web3Endpoint, null, alerts);
                            });
                            process_alerts();
                            return;
                        }
                        if ('undefined' === typeof data['id']) {
                            console.log('No "id" field in the data returned for network', web3Endpoint);
                            window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].forEach(function(callback) {
                                callback.call(null, 'No "id" field in the data returned for network ' + web3Endpoint, null, alerts);
                            });
                            process_alerts();
                            return;
                        }
                        if (1 !== parseInt(data['id'])) {
                            console.log('Wrong "id" field in the data returned for network', web3Endpoint);
                            window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].forEach(function(callback) {
                                callback.call(null, 'Wrong "id" field in the data returned for network ' + web3Endpoint, null, alerts);
                            });
                            process_alerts();
                            return;
                        }
                        if ('undefined' === typeof data['result']) {
                            console.log('No "result" field in the data returned for network', web3Endpoint);
                            window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].forEach(function(callback) {
                                callback.call(null, 'No "result" field in the data returned for network ' + web3Endpoint, null, alerts);
                            });
                            process_alerts();
                            return;
                        }
                        window.ethereumWallet._url_to_network_id_call_success = true;
                        const blockchainId = 'a' + data['result'];
                        if ('undefined' === typeof blockchainInfos[blockchainId]) {
                            console.log('Unsupported blockchainId(' + blockchainId + ') returned for', web3Endpoint, blockchainInfos);
                            window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].forEach(function(callback) {
                                callback.call(null, 'Unsupported blockchainId(' + blockchainId + ') returned for ' + web3Endpoint, null, alerts);
                            });
                            process_alerts();
                            return;
                        }
                        if ('undefined' === typeof window.ethereumWallet._url_to_network_id_cache[web3Endpoint]) {
                            window.ethereumWallet._url_to_network_id_cache[web3Endpoint] = blockchainId;
                        }
                        window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].forEach(function(callback) {
                            callback.call(null, null, blockchainId, alerts);
                        });
                        process_alerts();
                    },
                    // @see https://stackoverflow.com/a/37311074/4256005
                    error: function(xhr, status, error) {
                        window.ethereumWallet._url_to_network_id_call_success = '<?php 
    echo  $failedNetworkIdRequestMessage ;
    ?>';
                        // handle error
                        console.log([xhr.status, xhr.responseText, status, error]);
                        window.ethereumWallet._url_to_network_id_call_started[web3Endpoint].forEach(function(callback) {
                            callback.call(null, '<?php 
    echo  $failedNetworkIdRequestMessage ;
    ?>', null, alerts);
                        });
                        process_alerts();
                    }
                })
            }, 1000);
        }

        window.ethereumWallet.ethereum_wallet_getBalance = function(web3Endpoint, callback) {
            let alerts = {};

            function process_alerts() {
                Object.keys(alerts).forEach(function(message) {
                    alert(message);
                });
            }

            if ('' == web3Endpoint) {
                callback.call(null, null, alerts);
                process_alerts();
                return;
            }
            const dataRequest = '{"jsonrpc":"2.0","method":"eth_getBalance","params":["0x476bb28bc6d0e9de04db5e19912c392f9a76535d", "latest"],"id":1}';

            function processData(data) {
                // data: {"jsonrpc":"2.0","id":1,"result":"0x28bf65e9896a69840"}
                if (!data) {
                    console.log('Empty data returned for network', web3Endpoint);
                    callback.call(null, 'Empty data returned for network ' + web3Endpoint, alerts);
                    process_alerts();
                    return;
                }
                if ('undefined' === typeof data['id']) {
                    console.log('No "id" field in the data returned for network', web3Endpoint);
                    callback.call(null, 'No "id" field in the data returned for network ' + web3Endpoint, alerts);
                    process_alerts();
                    return;
                }
                if (1 !== parseInt(data['id'])) {
                    console.log('Wrong "id" field in the data returned for network', web3Endpoint);
                    callback.call(null, 'Wrong "id" field in the data returned for network ' + web3Endpoint, alerts);
                    process_alerts();
                    return;
                }
                if ('undefined' !== typeof data['error']) {
                    if ('undefined' === typeof data['error']['message']) {
                        console.log('No "message" field in the error data returned for network', web3Endpoint);
                        callback.call(null, 'No "message" field in the error data returned for network ' + web3Endpoint, alerts);
                        process_alerts();
                        return;
                    }
                    console.log('The error message returned for network', web3Endpoint, ':', data['error']['message']);
                    callback.call(null, data['error']['message'], alerts);
                    process_alerts();
                    return;
                }
                if ('undefined' === typeof data['result']) {
                    console.log('No "result" field in the data returned for network', web3Endpoint);
                    callback.call(null, 'No "result" field in the data returned for network ' + web3Endpoint, alerts);
                    process_alerts();
                    return;
                }
                callback.call(null, null, alerts);
                process_alerts();
            }
            jQuery.ajax({
                url: web3Endpoint,
                type: "POST",
                data: dataRequest,
                contentType: "application/json; charset=utf-8",
                dataType: "json",
                success: processData,
                // @see https://stackoverflow.com/a/37311074/4256005
                error: function(xhr, status, error) {
                    // handle error
                    console.log([xhr.status, xhr.responseText, status, error]);
                    if ('' !== xhr.responseText) {
                        const data = JSON.parse(xhr.responseText);
                        processData(data);
                        return
                    }
                    callback.call(null, 'Failed to check selected infura.io blockchain availability', alerts);
                    process_alerts();
                }
            })
        }

        jQuery(document).ready(function() {
            const blockchainInfos = JSON.parse('<?php 
    echo  json_encode( $blockchainInfos ) ;
    ?>');

            function ethereum_wallet_blockchain_getWeb3Endpoint() {
                const blockchain_api_node = jQuery('#ETHEREUM_WALLET_blockchain_api_node').find(":selected").val();
                switch (blockchain_api_node) {
                    case 'infuraio':
                        const infuraApiKey = jQuery('#ETHEREUM_WALLET_infuraApiKey').val();
                        const blockchain = jQuery('#ETHEREUM_WALLET_blockchain_network').find(":selected").val();
                        return "https://" + blockchain + ".infura.io/v3/" + infuraApiKey;
                    case 'custom':
                        return jQuery('#ETHEREUM_WALLET_web3Endpoint').val();
                }
                return null;
            }
            window.ethereumWallet.ethereum_wallet_blockchain_getWeb3Endpoint = ethereum_wallet_blockchain_getWeb3Endpoint;

            function ethereum_wallet_blockchain_adjust_fields(blockchainId) {
                let p = jQuery('#ETHEREUM_WALLET_etherscanApiKey').parent().find('p:first');
                let asterix = jQuery('#ETHEREUM_WALLET_etherscanApiKey').parent().parent().parent().parent().find('th').find('sup');
                if (null === blockchainId || 'undefined' === typeof blockchainInfos[blockchainId]) {
                    p.hide();
                    asterix.hide();
                    return;
                }
                const blockchainInfo = blockchainInfos[blockchainId];
                const template = '<?php 
    echo  $template ;
    ?>';
                const arg1 = '<a target="_blank" href="' + blockchainInfo['explorer_api_url'] + '">' + blockchainInfo['base_explorer_display_url'] + '</a>';
                const arg2 = '<a target="_blank" href="' + blockchainInfo['explorer_register_url'] + '">';
                const arg3 = '</a>';
                const description = template.replace('%1$s', arg1).replace('%2$s', arg2).replace('%3$s', arg3);
                p.html(description);
                if (null === blockchainInfo['explorer_api_url']) {
                    p.hide();
                    asterix.hide();
                } else {
                    p.show();
                    asterix.show();
                }
            }
            window.ethereumWallet.ethereum_wallet_blockchain_adjust_fields = ethereum_wallet_blockchain_adjust_fields;

            jQuery('#ETHEREUM_WALLET_blockchain_network').on('change', function(e) {
                const blockchain = jQuery(this).find(":selected").val();
                ethereum_wallet_blockchain_adjust_fields(blockchain);
                const web3Endpoint = ethereum_wallet_blockchain_getWeb3Endpoint();
                window.ethereumWallet.ethereum_wallet_getBalance.call(null, web3Endpoint, function(error, alerts) {
                    if (null !== error) {
                        ethereum_wallet_blockchain_adjust_fields(null);
                        if (-1 !== error.indexOf('project ID does not have access')) {
                            error = '<?php 
    _e( 'Premium Infura.io account is required to use this network with the Infura.io service', 'ethereum-wallet' );
    ?>';
                        }
                        alerts[error] = 1;
                        window.ethereumWallet._url_to_network_id_call_success = error;
                        return;
                    }
                    window.ethereumWallet._url_to_network_id_call_success = true;
                });
            });
            <?php 
    if ( 'infuraio' === $blockchain_api_node_current ) {
        ?>
                jQuery('#ETHEREUM_WALLET_blockchain_network').change();
            <?php 
    }
    ?>

            function ethereum_wallet_refresh_blockchain(e) {
                const web3Endpoint = jQuery('#ETHEREUM_WALLET_web3Endpoint').val();
                window.ethereumWallet.ethereum_wallet_get_blockchain_network_id.call(null, web3Endpoint, blockchainInfos, function(error, blockchainId, alerts) {
                    if (null !== error) {
                        ethereum_wallet_blockchain_adjust_fields(null);
                        alerts['<?php 
    echo  $failedNetworkIdRequestMessage ;
    ?>'] = 1;
                        return;
                    }
                    ethereum_wallet_blockchain_adjust_fields(blockchainId);
                });
            }
            jQuery('#ETHEREUM_WALLET_web3Endpoint').on('change', ethereum_wallet_refresh_blockchain);
            <?php 
    if ( 'custom' === $blockchain_api_node_current ) {
        ?>
                jQuery('#ETHEREUM_WALLET_web3Endpoint').change();
            <?php 
    }
    ?>
        });
    </script>

    <tr valign="top" class="ethereum-wallet-blockchain-api-node ethereum-wallet-blockchain-api-node-custom">
        <th scope="row" colspan="2">
            <h2><?php 
    _e( "Blockchain history and tx sending", 'ethereum-wallet' );
    ?></h2>
        </th>
    </tr>

    <tr valign="top">
        <th scope="row"><?php 
    _e( "Blockchain Explorer Api Key", 'ethereum-wallet' );
    ?>
            <?php 
    if ( !is_null( $explorer_api_url ) ) {
        ?>
                <sup>*</sup>
            <?php 
    }
    ?>
        </th>
        <td>
            <fieldset>
                <label>
                    <input class="text" name="ETHEREUM_WALLET_etherscanApiKey" id="ETHEREUM_WALLET_etherscanApiKey" type="text" maxlength="35" placeholder="<?php 
    _e( "Put your Blockchain Explorer (like etherscan.io) Api Key there", 'ethereum-wallet' );
    ?>" value="<?php 
    echo  ( !empty($options['etherscanApiKey']) ? esc_attr( $options['etherscanApiKey'] ) : '' ) ;
    ?>">
                    <p><?php 
    echo  sprintf(
        $template,
        '<a target="_blank" href="' . $explorer_api_url . '">' . $base_explorer_display_url . '</a>',
        '<a target="_blank" href="' . $explorer_register_url . '">',
        '</a>'
    ) ;
    ?></p>
                    <p><?php 
    _e( "Required for transactions history display and any ERC20 and NFT tokens related functionality.", 'ethereum-wallet' );
    ?></p>
                </label>
            </fieldset>
        </td>
    </tr>

    <tr valign="top">
        <th scope="row"><?php 
    _e( "Gas Limit", 'ethereum-wallet' );
    ?></th>
        <td>
            <fieldset>
                <label>
                    <input class="text" name="ETHEREUM_WALLET_gas_limit" type="number" min="0" step="10000" maxlength="8" placeholder="200000" value="<?php 
    echo  ( !empty($options['gas_limit']) ? esc_attr( $options['gas_limit'] ) : '200000' ) ;
    ?>">
                    <p><?php 
    _e( "The default gas limit to to spend on your transactions. 200000 is a reasonable default value.", 'ethereum-wallet' );
    ?></p>
                </label>
            </fieldset>
        </td>
    </tr>

    <tr valign="top">
        <th scope="row"><?php 
    _e( "Gas price", 'ethereum-wallet' );
    ?></th>
        <td>
            <fieldset>
                <label>
                    <input class="text" name="ETHEREUM_WALLET_gas_price" type="number" min="0" step="1" maxlength="8" placeholder="200" value="<?php 
    echo  ( !empty($options['gas_price']) ? esc_attr( $options['gas_price'] ) : '200' ) ;
    ?>">
                    <p><?php 
    _e( "The gas price in Gwei. Reasonable values are in a 50-250 ratio. The default value is 200 to ensure that your tx would be sent in most of the time.", 'ethereum-wallet' );
    ?></p>
                    <p><?php 
    _e( "The actual gas price used would be this value or less, depending on the current reasonable gas price in the blockchain.", 'ethereum-wallet' );
    ?></p>
                </label>
            </fieldset>
        </td>
    </tr>

<?php 
}

include_once $ETHEREUM_WALLET_plugin_dir . '/settings/advanced_blockchain.php';